from collections import Counter, defaultdict

import stim


def reduce_circuit_depth(circuit: stim.Circuit):
    """Apply basic circuit depth reduction by accumulating commuting gates.

    Currently supports I, S, Z, S_DAG, CZ, H, CX,
    TICK, and repeat blocks.
    Accumulates the S, Z, S_DAG, and CZ gates
    and removes the I gates.

    :param circuit: Input circuit.
    :returns reduced_circuit: An equivalent circuit of reduced depth.
    :raises ValueError: If an unsupported instruction is encountered.
    """
    return _DepthReducer.reduce(circuit)


class _DepthReducer:

    _GATE_TO_EXPONENT = {'S': 1, 'Z': 2, 'S_DAG': 3}
    _PERIOD = 4

    def __init__(self) -> None:
        self.s_gates: Counter[int] = Counter()
        """Map from qubit index to the integer exponent of the S gate applied."""
        self.cz_to_moment: dict[frozenset[int], int] = {}
        """Map from an unordered pair of qubit indices to the timeslice of its CZ gate."""
        self.moment_to_cz: defaultdict[int, dict[int, int]] = defaultdict(dict)
        """Map from timeslice to a bidirectional map between qubit indices."""
    
    @classmethod
    def reduce(cls, circuit: stim.Circuit) -> stim.Circuit:
        """See `reduce_circuit_depth`."""
        out = stim.Circuit()
        reducer = cls()
        for instruction in circuit:
            if isinstance(instruction, stim.CircuitRepeatBlock):
                out += reducer._to_circuit()
                reducer = cls()
                out += instruction.repeat_count * cls.reduce(instruction.body_copy())
            elif instruction.name in {'TICK', 'H', 'CX'}:
                out += reducer._to_circuit()
                reducer = cls()
                out.append(instruction)
            elif instruction.name == 'CZ':
                reducer._add_cz_gates(instruction.target_groups())
            elif instruction.name in cls._GATE_TO_EXPONENT:
                reducer._add_1_qubit_gates(
                    name=instruction.name,
                    targets=instruction.targets_copy(),
                )
            elif instruction.name != 'I':
                raise ValueError(f"Unsupported instruction {instruction.name} in circuit.")
        return out + reducer._to_circuit()

    def _to_circuit(self):
        """Materialize the currently accumulated gates as a circuit.

        :returns circuit: A ``stim.Circuit`` containing the gates accumulated in ``self``.
        """
        out = stim.Circuit()
        
        _1_qubit_gates: dict[int, set[int]] = defaultdict(set)
        for qubit_index, exponent in self.s_gates.items():
            _1_qubit_gates[exponent % self._PERIOD].add(qubit_index)
        for name, exponent in self._GATE_TO_EXPONENT.items():
            if targets := _1_qubit_gates[exponent]:
                out.append(name, targets, ())
        
        _max_moment = max(self.moment_to_cz.keys(), default=-1)
        for t in range(_max_moment + 1):
            for index_0, index_1 in self.moment_to_cz[t].items():
                if index_0 < index_1:
                    out.append("CZ", [index_0, index_1], ())
        
        return out

    def _add_1_qubit_gates(
            self,
            name: str,
            targets: list[stim.GateTarget],
    ):
        exponent = self._GATE_TO_EXPONENT[name]
        for target in targets:
            self.s_gates[target.value] += exponent

    def _add_cz_gates(self, target_groups: list[list[stim.GateTarget]]):
        for target_0, target_1 in target_groups:
            index_0, index_1 = target_0.value, target_1.value
            pair = frozenset({index_0, index_1})
            if pair in self.cz_to_moment.keys():
                t = self.cz_to_moment.pop(pair)
                self.moment_to_cz[t].pop(index_0)
                self.moment_to_cz[t].pop(index_1)
            else:
                t = self._get_earliest_moment(pair)
                self.cz_to_moment[pair] = t
                self.moment_to_cz[t][index_0] = index_1
                self.moment_to_cz[t][index_1] = index_0

    def _get_earliest_moment(self, pair: frozenset[int]):
        """Get the earliest timeslice when a specified CZ can be scheduled.

        :param pair: The unordered pair of physical qubit indices
            that the CZ gate should act upon.
        :returns moment: The earliest moment index that does not conflict.
        """
        t = 0
        while any(index in self.moment_to_cz[t] for index in pair):
            t += 1
        return t


from collections.abc import Collection, Iterable, Sequence
import itertools
from functools import cached_property
import math
from typing import Literal

import numpy as np
import stim

from qrmfold import logical_qubit_orderings
from qrmfold._automorphism import Automorphism
from qrmfold._depth_reducer import reduce_circuit_depth
from qrmfold.utils import all_bitstrings, complement, extract_arguments, powerset, sign_to_power, rref_gf2


def _multiply(*vectors: Iterable[int]):
    """Elementwise multiplication of vectors.
    
    Require at least one vector, and all vectors must have the same length.
    """
    return [math.prod(a) for a in zip(*vectors, strict=True)]

def _flip(vector: Iterable[int]):
    """Subtract a vector over GF(2) from 1."""
    return [1 - a for a in vector]


class ReedMuller:
    """The classical Reed-Muller code RM(r, m)."""

    def __init__(self, order: int, bit_count: int, minimize_weight: bool = True):
        """
        :param order: The order r of the Reed-Muller code.
        :param bit_count: The number m of bits of the Reed-Muller code.
        :param minimize_weight: Whether to minimize the weight
            of each generator to 2^{m/2 + 1}.
        """
        self.ORDER = order
        """The order r of the Reed-Muller code."""
        self.BIT_COUNT = bit_count
        """The number m of bits of the Reed-Muller code."""
        self.BASIS = tuple(self._construct_basis_vector(i) for i in range(bit_count))
        """Basis vectors {v_i : 0 <= i < m} for the Reed-Muller code of m bits."""

        _all_ones = [1] * 2**bit_count
        _generator_matrix: list[list[int]] = []
        _basis_neg = tuple(_flip(v) for v in self.BASIS)
        for cardinality in range(order + 1):
            for subset in itertools.combinations(range(bit_count), cardinality):
                vector_subset = [self.BASIS[i] for i in subset]
                generator: list[int] = _all_ones if not subset else _multiply(*vector_subset)
                
                if minimize_weight:
                    c = bit_count//2 - 1 - cardinality
                    _complement = complement(bit_count, subset, start=0)
                    additional_vectors = [_basis_neg[i] for i in _complement][:c]
                    generator = _multiply(generator, *additional_vectors)
                
                _generator_matrix.append(generator)
        self.generator_matrix = _generator_matrix
        """Matrix with a row for every subset of the basis vectors
        of cardinality at most the order r of the code.
        """

    def _construct_basis_vector(self, index: int):
        """Construct the ith basis vector of the Reed-Muller code.

        The vector is a string of 2^i zeros followed by 2^i ones repeated 2^(m - i - 1) times.

        :param index: Basis index i satisfying 0 <= i < m,
        where m is the bit count of the Reed-Muller code.
        :returns basis_vector: The ith basis vector of length 2^m.
        """
        list_: list[int] = ([0] * 2**index + [1] * 2**index) * 2 ** (self.BIT_COUNT-index-1)
        return tuple(list_)


def _get_residual_slices(bit_count: int, i: int, j: int):
    """Generate residual bitstring slices used by the automorphisms.

    :param bit_count: Bit count of the bit index.
    :param i: First basis vector index (1-indexed).
    :param j: Second basis vector index (1-indexed).
    :returns residual_slices: An iterator over triples of residual bitstrings.
    """
    smaller, larger = min(i, j), max(i, j)
    residual_1 = all_bitstrings(bit_count-larger)
    residual_2 = all_bitstrings(larger-1-smaller)
    residual_3 = all_bitstrings(smaller-1)
    return itertools.product(residual_1, residual_2, residual_3)


def _p_automorphism(bit_count: int, i: int, j: int):
    """Swap basis vectors v_i and v_j (1-indexed).

    :param bit_count: Bit count of the bit index.
    :param i: First basis vector index (1-indexed).
    :param j: Second basis vector index (1-indexed).
    :returns pairs: A list of swapped bit index pairs ``(s1, s2)``.
        For each pair, ``s1`` has a 0 in position ``i`` and a 1 in position ``j``,
        and ``s2`` has a 1 in position ``i`` and a 0 in position ``j``.
    """
    result: list[tuple[str, str]] = []
    for a, b, c in _get_residual_slices(bit_count, i, j):
        s1 = a + '0' + b + '1' + c
        s2 = a + '1' + b + '0' + c
        result.append((s1, s2))
    return result


def _q_automorphism(bit_count: int, i: int, j: int):
    """Add basis vector v_j onto basis vector v_i (1-indexed).

    :param bit_count: Bit count of the bit index.
    :param i: Target basis vector index (1-indexed).
    :param j: Source basis vector index (1-indexed).
    :returns pairs: A list of swapped bit index pairs ``(s1, s2)``.
        For each pair, ``s1`` has a 0 in position ``i`` and a 1 in position ``j``,
        and ``s2`` has a 1 in position ``i`` and a 1 in position ``j``.
    """
    if i < j:
        s1_larger, s1_smaller = '1', '0'
    else:
        s1_larger, s1_smaller = '0', '1'
    result: list[tuple[str, str]] = []
    for a, b, c in _get_residual_slices(bit_count, i, j):
        s1 = a + s1_larger + b + s1_smaller + c
        s2 = a + '1' + b + '1' + c
        result.append((s1, s2))
    return result


def _get_intermediate_subsets(b_subset: set[int], b_prime_subset: set[int]):
    """Compute intermediate subsets between two subsets B and B'.

    :param b_subset: Starting subset B.
    :param b_prime_subset: Ending subset B'.
    :returns intermediate_subsets: A list such that each subset in
        ``(b_subset, *intermediate_subsets, b_prime_subset)``
        differs from the previous one by exactly one element.
    :raises ValueError: If B and B' do not differ by
        at least one element.
    """
    b_take_b_prime = b_subset.difference(b_prime_subset)
    b_prime_take_b = b_prime_subset.difference(b_subset)
    try:
        b_take_b_prime.pop()
        b_prime_take_b.pop()
    except KeyError:
        raise ValueError("b_subset and b_prime_subset must differ by at least one element.")
    intermediate_subsets: list[set[int]] = []
    current_subset = b_subset.copy()
    for i, j in zip(b_take_b_prime, b_prime_take_b, strict=True):
        current_subset.remove(i)
        current_subset.add(j)
        intermediate_subsets.append(current_subset.copy())
    return intermediate_subsets


_signature_to_pauli = {
    (False, False): '_',
    (False, True): 'X',
    (True, True): 'Y',
    (True, False): 'Z',
}
"""Map from anticommutation with (X, Z) to the unique Pauli operator with that signature."""


class QuantumReedMuller:
    """The quantum Reed-Muller code CSS(RM(m/2 - 1, m), RM(m/2 - 1, m))."""

    def __init__(
            self,
            bit_count: int,
            minimize_weight: bool = True,
            logical_qubit_ordering: None | dict[int, set[int]] = None,
    ):
        """
        :param bit_count: The number m of bits of the
            underlying classical Reed-Muller code; must be even.
        :param minimize_weight: Whether to minimize the weight
            of each stabilizer generator to 2^{m/2 + 1}.
        :param logical_qubit_ordering: Optional 1-to-1 map from logical qubit
            index to a subset of [m] of cardinality m/2.
            If not specified, subsets are ordered lexicographically.
        :raises ValueError: If ``bit_count`` is odd,
        or if ``logical_qubit_ordering`` does not have the required codomain.
        """
        self._validate_init_inputs(bit_count, logical_qubit_ordering)
        self.BIT_COUNT = bit_count
        """The number m of bits of the underlying classical Reed-Muller code."""
        order = bit_count//2 - 1
        self.classical = ReedMuller(order, bit_count, minimize_weight=minimize_weight)
        """The underlying classical Reed-Muller code RM(m/2 - 1, m)."""
        self.stabilizer_generators = {
            pauli: [stim.PauliString(pauli if p else 'I' for p in g) for g in self.classical.generator_matrix]
            for pauli in ('X', 'Z')
        }
        """Map from Pauli (X or Z) to a list of stabilizer generators of that type."""
        self._logical_x_supports: dict[frozenset[int], list[int]] = {
            frozenset(subset): _multiply(*(self.classical.BASIS[i-1] for i in subset))
            for subset in itertools.combinations(range(1, bit_count+1), bit_count//2)
        }
        self.logical_operators: dict[frozenset[int], tuple[stim.PauliString, stim.PauliString]] = {}
        """Map from logical qubit (labelled by a subset of [m]) to its X and Z logical operators."""
        for subset, support in self._logical_x_supports.items():
            x = stim.PauliString('X' if p else 'I' for p in support)
            _complement = frozenset(complement(bit_count, subset))
            z = stim.PauliString('Z' if p else 'I' for p in self._logical_x_supports[_complement])
            self.logical_operators[subset] = (x, z)
        self.logical_qubit_ordering = logical_qubit_orderings.lexicographic(bit_count) \
            if logical_qubit_ordering is None else logical_qubit_ordering
        """Map from logical qubit index to a unique subset of [m] of cardinality m/2."""
        self.subset_to_logical_index = {
            frozenset(subset): index for index, subset in self.logical_qubit_ordering.items()
        }
        """Map from subset of [m] of cardinality m/2 to logical qubit index."""

    @staticmethod
    def _validate_init_inputs(bit_count: int, logical_qubit_ordering: None | dict[int, set[int]]):
        """Validate constructor inputs.

        :param bit_count: The number m of bits of the
            underlying classical Reed-Muller code; must be even.
        :param logical_qubit_ordering: Candidate map from logical index to subset of [m] of cardinality m/2.
        :raises ValueError: If ``bit_count`` is odd,
        or if ``logical_qubit_ordering`` does not have the required codomain.
        """
        if bit_count % 2:
            raise ValueError("bit_count must be even")
        if logical_qubit_ordering is None:
            return
        target: set[frozenset[int]] = {frozenset(subset) for subset in itertools.combinations(range(1, bit_count+1), bit_count//2)}
        actual: set[frozenset[int]] = {frozenset(subset) for subset in logical_qubit_ordering.values()}
        if target != actual:
            raise ValueError("logical_qubit_ordering must be a 1-to-1 map from logical qubit index to subset of [m] of cardinality m/2.")

    def print(self):
        """Print the stabilizer generators and logical operators."""
        for pauli in ('X', 'Z'):
            print(f'{pauli} stabilizer generators:')
            for g in self.stabilizer_generators[pauli]:
                print(g)
        print('logical qubits and their X and Z operators:')
        digit_count = math.ceil(math.log10(len(self._logical_x_supports)))
        for logical_index, subset in self.logical_qubit_ordering.items():
            print(
                f'{logical_index:{digit_count}d}',
                subset,
                *(str(pauli) for pauli in self.logical_operators[frozenset(subset)]),
            )

    @cached_property
    def stabilizer_generators_rref(self):
        """The stabilizer generators as a binary symplectic matrix in reduced row echelon form."""
        stabilizers_bsf: list[np.ndarray[tuple[int], np.dtype[np.bool_]]] = []
        for basis_generators in self.stabilizer_generators.values():
            for generator in basis_generators:
                xs, zs = generator.to_numpy()
                stabilizers_bsf.append(np.append(xs, zs))
        return rref_gf2(stabilizers_bsf)

    def _automorphism(
            self,
            automorphism_type: Literal['trivial', 'P', 'Q'] = 'trivial',
            pairs: None | Iterable[tuple[int, int]] = None,
    ):
        """Construct an automorphism of the underlying classical Reed-Muller code.

        :param automorphism_type: The automorphism type: trivial, P, or Q.
            The trivial automorphism maps every element to itself.
            P(i, j) swaps basis vectors v_i and v_j.
            Q(i, j) adds basis vector v_j onto basis vector v_i.
        :param pairs: Iterable of integer pairs (i, j) satisfying
            1 <= i, j <= m, where m is the bit count of the quantum Reed-Muller code.
            Each integer in ``pairs`` must be distinct.
            If omitted (or if ``automorphism_type == 'trivial'``),
            the trivial automorphism is returned.
        :returns automorphism: A :class:`qrmfold._automorphism.Automorphism` object.
        :raises ValueError: If any index in ``pairs`` is out of range or is repeated,
            or if ``automorphism_type`` is not recognized.
        """
        automorphism = Automorphism(self.BIT_COUNT, ())
        if pairs is not None and automorphism_type != 'trivial':
            exhausted_indices: set[int] = set()
            for i, j in pairs:
                if not ((1 <= i <= self.BIT_COUNT) and (1 <= j <= self.BIT_COUNT)):
                    raise ValueError(f"i and j must be between 1 and {self.BIT_COUNT}, inclusive")
                if (pair:={i, j}) & exhausted_indices:
                    raise ValueError("Each integer in pairs must be distinct")
                exhausted_indices.update(pair)
                if automorphism_type == 'P':
                    automorphism.update(_p_automorphism(self.BIT_COUNT, i, j))
                elif automorphism_type == 'Q':
                    automorphism.update(_q_automorphism(self.BIT_COUNT, i, j))
                else:
                    raise ValueError("automorphism_type must be 'trivial', 'P', or 'Q'")
        return automorphism

    def automorphism(
            self,
            pairs: None | Collection[tuple[int, int]] = None,
            automorphism_type: Literal['trivial', 'P', 'Q'] = 'Q',
            gate_type: Literal['swap', 'phase'] = 'phase',
    ):
        """Return the physical circuit of U_t(a(K)).

        :param pairs: Set K of integer pairs (i, j) satisfying
            1 <= i, j <= m, where m is the bit count of the quantum Reed-Muller code.
            Each integer in ``pairs`` must be distinct.
            If omitted, treated as empty.
        :param automorphism_type: The automorphism type a: trivial, P, or Q.
            The trivial automorphism maps every element to itself.
            P(i, j) swaps basis vectors v_i and v_j.
            Q(i, j) adds basis vector v_j onto basis vector v_i.
        :param gate_type: Gate type t: swap or phase.
        :returns circuit: A ``stim.Circuit`` representing the physical circuit of U_t(a(K)).
        """
        return self._automorphism(automorphism_type=automorphism_type, pairs=pairs).gate(gate_type)
    
    def automorphism_product(
            self,
            pairs: None | Collection[tuple[int, int]] = None,
            automorphism_type: Literal['trivial', 'P', 'Q'] = 'Q',
            gate_type: Literal['swap', 'phase'] = 'phase',
    ):
        """Return the physical circuit of ``\\prod_{L \\subseteq K} U_t(a(L))``.

        :param pairs: Set K of integer pairs (i, j) satisfying
            1 <= i, j <= m, where m is the bit count of the quantum Reed-Muller code.
            Each integer in ``pairs`` must be distinct.
            If omitted, treated as empty.
        :param automorphism_type: The automorphism type a: trivial, P, or Q.
            The trivial automorphism maps every element to itself.
            P(i, j) swaps basis vectors v_i and v_j.
            Q(i, j) adds basis vector v_j onto basis vector v_i.
        :param gate_type: Gate type t: swap or phase.
        :returns circuit: A ``stim.Circuit`` representing the physical circuit of the product.
        """
        if pairs is None:
            pairs = []
        return sum((self.automorphism(
            pairs=l_subset,
            automorphism_type=automorphism_type,
            gate_type=gate_type,
        ) for l_subset in powerset(pairs)), start=stim.Circuit())

    def q_phase_logical_action(self, pairs: Collection[tuple[int, int]]):
        """Compute the logical action of U_P(Q(K)).

        :param pairs: Set K of pairs.
        :returns circuit: A ``stim.Circuit`` acting on the logical qubits of the code.
        """
        circuit = self._logical_action_starter()
        for l_subset in powerset(pairs, self.BIT_COUNT//2 - 2):
            self._logical_action_helper(circuit, l_subset, gates=['CZ'])
        if len(pairs) >= self.BIT_COUNT//2 - 1:
            for l_subset in itertools.combinations(pairs, self.BIT_COUNT//2 - 1):
                self._logical_action_helper(circuit, l_subset, gates=['CZ', 'Z'])
        if len(pairs) == self.BIT_COUNT//2:
            arguments_0 = extract_arguments(0, pairs)
            logical_index = self.subset_to_logical_index[frozenset(arguments_0)]
            gate = 'S_DAG' if self.BIT_COUNT//2 % 2 else 'S'
            circuit.append(gate, logical_index, ())
        return circuit

    def q_phase_product_logical_action(self, pairs: Collection[tuple[int, int]]):
        """Compute the logical action of ``\\prod_{L \\subseteq K} U_P(Q(L))``.

        :param pairs: Set K of pairs.
        :returns circuit: A ``stim.Circuit`` acting on the logical qubits of the code.
        :raises ValueError: If ``pairs`` longer than ``self.BIT_COUNT/2``.
        """
        circuit = self._logical_action_starter()
        if len(pairs) <= self.BIT_COUNT//2 - 2:
            self._logical_action_helper(circuit, pairs, gates=['CZ'])
        elif len(pairs) == self.BIT_COUNT//2 - 1:
            self._logical_action_helper(circuit, pairs, gates=['CZ', 'Z'])
        elif len(pairs) == self.BIT_COUNT//2:
            arguments_0 = extract_arguments(0, pairs)
            logical_index = self.subset_to_logical_index[frozenset(arguments_0)]
            gate = 'S_DAG' if self.BIT_COUNT//2 % 2 else 'S'
            circuit.append(gate, logical_index, ())
        else:
            raise ValueError("pairs cannot be longer than self.BIT_COUNT/2")
        return circuit

    def _logical_action_starter(self):
        circuit = stim.Circuit()
        circuit.append('I', self.logical_qubit_ordering.keys(), ())
        return circuit

    def _logical_action_helper(
            self,
            circuit: stim.Circuit,
            pairs: Collection[tuple[int, int]],
            gates: Sequence[str],
    ):
        """Helper for :meth:`q_phase_logical_action` and :meth:`q_phase_product_logical_action`.

        :param circuit: Circuit to append to.
        :param pairs: Pair set (see caller) of length <m/2.
        :param gates: Gate names to append for each affected logical qubit pair.
        :returns None: Mutates ``circuit`` in-place.
        """
        arguments_0 = extract_arguments(0, pairs)
        arguments_1 = extract_arguments(1, pairs)
        encountered_qubits: set[int] = set()
        for logical_index, b_subset in self.logical_qubit_ordering.items():
            if logical_index not in encountered_qubits and arguments_0.issubset(b_subset) and arguments_1.isdisjoint(b_subset):
                b_complement = complement(self.BIT_COUNT, b_subset)
                b_prime_subset = b_complement.union(arguments_0).difference(arguments_1)
                logical_index_prime = self.subset_to_logical_index[frozenset(b_prime_subset)]
                for gate in gates:
                    circuit.append(gate, [logical_index, logical_index_prime], ())
                encountered_qubits.add(logical_index_prime)

    def gate(
            self,
            name: Literal['S', 'H', 'ZZCZ', 'SWAP'],
            targets: Iterable[int],
            reduce_depth: bool = True,
    ):
        """Build the physical circuit inducing a logical gate.

        :param name: The name of the logical gate to implement.
            This can be S, H, ZZCZ, or SWAP.
        :param targets: Logical qubit indices to apply the gate on.
            For 2-qubit gates, the operation is broadcast over
            consecutive target pairs, so an even target count is required.
        :param reduce_depth: Whether to apply basic depth reduction before returning.
        :returns circuit: A ``stim.Circuit`` inducing the requested logical action.
        :raises ValueError: If a 2-qubit gate is requested with an odd number of targets.
        """
        if name == 'S' or name == 'H':
            _gate = self._s if name == 'S' else self._h
            out = sum((_gate(
                set(self.logical_qubit_ordering[target])
            ) for target in targets), start=stim.Circuit())
        else:
            out = stim.Circuit()
            try:
                for target_0, target_1 in itertools.batched(targets, 2, strict=True):
                    out += self._2_qubit_gate(target_0, target_1, name)
            except ValueError:
                raise ValueError(f"2-qubit gate {name} requires an even target count but was given {targets}.")
        if reduce_depth:
            return reduce_circuit_depth(out)
        return out

    def _s(self, b_subset: set[int]):
        """Build the physical circuit inducing a logical S.

        :param b_subset: Logical qubit label B as a subset of [m] of cardinality m/2.
        :returns circuit: A ``stim.Circuit`` inducing the logical S on that qubit.
        """
        pairs = list(zip(b_subset, complement(self.BIT_COUNT, b_subset), strict=True))
        out = self.automorphism_product(pairs, automorphism_type='Q', gate_type='phase')
        if self.BIT_COUNT//2 % 2:
            return out.inverse()
        return out

    def _h(self, b_subset: set[int]):
        """Build the physical circuit inducing a logical H.

        :param b_subset: Logical qubit label B as a subset of [m] of cardinality m/2.
        :returns circuit: A ``stim.Circuit`` inducing the logical H on that qubit.
        """
        s_b = self._s(b_subset)
        transversal_h = stim.Circuit(f'H {' '.join(str(k) for k in range(s_b.num_qubits))}')
        s_b_complement = self._s(complement(self.BIT_COUNT, b_subset))
        return s_b + transversal_h + s_b_complement + transversal_h + s_b
    
    def _2_qubit_gate(self, target_0: int, target_1: int, name: Literal['SWAP', 'ZZCZ']):
        """Build a physical circuit implementing a 2-qubit logical gate.

        :param target_0: First logical qubit index.
        :param target_1: Second logical qubit index.
        :param name: 2-qubit logical gate: SWAP or ZZCZ.
        :returns circuit: A ``stim.Circuit`` inducing the requested logical action.
        """
        _gate = self._swap_restricted if name == 'SWAP' else self._zzcz_restricted
        b_subset = set(self.logical_qubit_ordering[target_0])
        b_prime_subset = set(self.logical_qubit_ordering[target_1])
        intermediate_subsets = _get_intermediate_subsets(b_subset, b_prime_subset)
        if not intermediate_subsets:
            return _gate(b_subset, b_prime_subset)
        entry = sum((
            self._swap_restricted(a, b) for a, b in
            itertools.pairwise([b_subset] + intermediate_subsets)
        ), start=stim.Circuit())
        apex = _gate(intermediate_subsets[-1], b_prime_subset)
        return entry + apex + entry.inverse()

    def _zzcz_restricted(self, b_subset: set[int], b_prime_subset: set[int]):
        """Build the physical circuit inducing a logical (ZZ)CZ.

        The two logical qubits must be labelled by subsets that differ by
        exactly one element.

        :param b_subset: First logical label B as a subset of [m] of cardinality m/2.
        :param b_prime_subset: Second logical label B'.
        :returns circuit: A ``stim.Circuit`` implementing logical (ZZ)CZ on B and B'.
        :raises ValueError: If B and B' do not differ by exactly one element.
        """
        arguments_0 = b_subset.intersection(b_prime_subset)
        if len(arguments_0) != self.BIT_COUNT//2 - 1:
            raise ValueError(f"Logical qubit labels {b_subset}, {b_prime_subset} must differ by exactly one element.")
        arguments_1 = complement(self.BIT_COUNT, b_subset.union(b_prime_subset))
        pairs = list(zip(arguments_0, arguments_1, strict=True))
        return self.automorphism_product(pairs, automorphism_type='Q', gate_type='phase')

    def _swap_restricted(self, b_subset: set[int], b_prime_subset: set[int]):
        """Build the physical circuit inducing a logical SWAP.

        The two logical qubits must be labelled by subsets that differ by
        exactly one element.
        
        :param b_subset: First logical label B as a subset of [m] of cardinality m/2.
        :param b_prime_subset: Second logical label B'.
        :returns circuit: A ``stim.Circuit`` implementing logical SWAP on B and B'.
        :raises ValueError: If B and B' do not differ by exactly one element.
        """
        zzcz = self._zzcz_restricted(b_subset, b_prime_subset)
        hh = self._h(b_subset) + self._h(b_prime_subset)
        return 3 * (zzcz + hh)

    def _get_logical_tableau(self, physical_circuit: stim.Circuit):
        """Compute the logical tableau induced by a physical circuit.

        :param physical_circuit: A physical circuit that preserve the
            stabilizer group.
        :returns tableau: The induced logical tableau.
        """
        xs: list[stim.PauliString] = []
        zs: list[stim.PauliString] = []
        for _, subset in sorted(self.logical_qubit_ordering.items(), key=lambda item: item[0]):
            for logical_operator, conjugated_generator_list in zip(
                self.logical_operators[frozenset(subset)],
                (xs, zs),
                strict=True,
            ):
                transformed = logical_operator.after(physical_circuit)
                transformed_logical_action = ''.join(_signature_to_pauli[tuple(
                    not transformed.commutes(observable) for observable in self.logical_operators[frozenset(_subset)]
                )] for _, _subset in sorted(self.logical_qubit_ordering.items(), key=lambda item: item[0])) # type: ignore

                # construct logical representative
                logical_representative = stim.PauliString()
                for (_, _subset), pauli in zip(
                    sorted(self.logical_qubit_ordering.items(), key=lambda item: item[0]),
                    transformed_logical_action,
                    strict=True,
                ):
                    _x, _z = self.logical_operators[frozenset(_subset)]
                    if pauli == 'X':
                        logical_representative *= _x
                    elif pauli == 'Y':
                        logical_representative *= 1j * _x * _z
                    elif pauli == 'Z':
                        logical_representative *= _z

                phase_stabilizer = transformed * logical_representative
                # extract phase
                phase_exponent = (sign_to_power[phase_stabilizer.sign] + \
                    len(phase_stabilizer.pauli_indices('Y'))) % 4

                conjugated_generator_list.append(1j**phase_exponent * stim.PauliString(transformed_logical_action))
        tableau = stim.Tableau.from_conjugated_generators(xs=xs, zs=zs)
        return tableau


import itertools
import math

from qrmfold.utils import complement


def lexicographic(bit_count: int, start_index: int = 0):
    """Return the lexicographic logical qubit ordering.

    :param bit_count: The number m of bits of the
        underlying classical Reed-Muller code; must be even.
    :param start_index: Starting logical index.
    :returns ordering: A map from logical qubit index to
        a unique subset of [m] of cardinality m/2.
    """
    return dict(enumerate((
        set(subset) for subset in
        itertools.combinations(range(1, bit_count+1), bit_count//2)
    ), start=start_index))


def canonical(bit_count: int, start_index: int = 0):
    """Return the canonical logical qubit ordering.

    The first half of the logical qubits indices is lexicographically ordered.
    The second half is the complement of the first half.

    :param bit_count: The number m of bits of the
        underlying classical Reed-Muller code; must be even.
    :param start_index: Starting logical index.
    :returns ordering: A map from logical qubit index to
        a unique subset of [m] of cardinality m/2.
    """
    map_: dict[int, set[int]] = {}
    logical_qubit_count = math.comb(bit_count, bit_count//2)
    one_to_m = range(1, bit_count+1)
    for index, subset in zip(
        range(start_index, start_index + logical_qubit_count//2),
        itertools.combinations(one_to_m, bit_count//2),
    ):
        map_[index] = set(subset)
        map_[logical_qubit_count//2 + index] = complement(bit_count, subset)
    return map_


import pytest

from qrmfold.utils import all_bitstrings


class TestAllBitstrings:

    @pytest.mark.parametrize("m", range(1, 6))
    def test_correct_count(self, m):
        bitstrings = all_bitstrings(m)
        assert len(bitstrings) == 2**m


from collections.abc import Iterable
from typing import Literal
import itertools

import numpy as np
import stim


class Automorphism:
    """An automorphism of the classical Reed-Muller code RM(m/2-1, m)."""

    def __init__(self, bit_count: int, pairs: Iterable[tuple[str, str]]):
        """
        :param bit_count: The number m of bits of the Reed-Muller code.
        :param pairs: Iterable of distinct bit index pairs specifying swaps.
        """
        self.BIT_COUNT = bit_count
        """The number m of bits of the Reed-Muller code."""
        list_ = list(range(2**bit_count))
        for label_1, label_2 in pairs:
            p1 = int(label_1, 2)
            p2 = int(label_2, 2)
            list_[p1], list_[p2] = list_[p2], list_[p1]
        self.positions = list_
        """The permutation of the automorphism, represented as a length-2^m list
        such that ``positions[k]`` is the position of bit k after permutation.
        """

    def __str__(self):
        return f"{list(range(2**self.BIT_COUNT))} ->\n{self.positions}"

    def update(self, pairs: Iterable[tuple[str, str]]):
        list_ = self.positions
        for label_1, label_2 in pairs:
            p1 = int(label_1, 2)
            p2 = int(label_2, 2)
            list_[p1], list_[p2] = list_[p2], list_[p1]

    @property
    def matrix(self):
        """The 2^m x 2^m permutation matrix of the automorphism."""
        n = 2**self.BIT_COUNT
        out = np.zeros((n, n), dtype=np.uint8)
        for p, q in enumerate(self.positions):
            out[q, p] = 1
        return out

    def gate(self, gate_type: Literal['swap', 'phase']):
        """Return the physical circuit of this automorphism.

        :param gate_type: Gate type i.e. swap or phase.
        :returns circuit: A ``stim.Circuit`` of this automorphism.
        :raises ValueError: If ``gate_type`` is not recognized.
        """
        out = stim.Circuit()
        if gate_type == 'swap':
            swap_gates = self._swap_type()
            out.append("SWAP", itertools.chain.from_iterable(swap_gates), ())
        elif gate_type == 'phase':
            cz_gates, s_gates = self._phase_type()
            out.append("CZ", itertools.chain.from_iterable(cz_gates), ())
            out.append("S", s_gates, ())
        else:
            raise ValueError(f"Unknown gate type: {gate_type}")
        return out

    def _swap_type(self):
        """Return the SWAP-type gate of this automorphism.

        :returns swap_gates: A set of unordered pairs,
        each representing a SWAP gate between physical two qubits.
        """
        swap_gates: set[frozenset[int]] = set()
        for row_index, position in enumerate(self.positions):
            if row_index < position:
                swap_gates.add(frozenset({row_index, position}))
        return swap_gates

    def _phase_type(self):
        """Return the phase-type gate of this automorphism.

        :returns cz_gates: A set of unordered pairs representing physical CZ gates.
        :returns s_gates: A set of physical qubit indices where an S gate is applied.
        """
        cz_gates: set[frozenset[int]] = set()
        s_gates: set[int] = set()
        for row_index, position in enumerate(self.positions):
            if row_index == position:
                s_gates.add(row_index)
            elif row_index < position:
                cz_gates.add(frozenset({row_index, position}))
        return cz_gates, s_gates


import itertools

import pytest

from qrmfold.quantum_reed_muller import _get_intermediate_subsets
from qrmfold.quantum_reed_muller import ReedMuller, QuantumReedMuller, _p_automorphism, _q_automorphism


@pytest.mark.parametrize("m", range(2, 12, 2))
def test_minimize_weight(m: int):
    for r in range(1, m//2):
        rm = ReedMuller(r, m, minimize_weight=True)
        for row in rm.generator_matrix:
            assert sum(row) == 2**(m//2 + 1)


def test_p_automorphism():
    swaps = _p_automorphism(4, 1, 2)
    expected_swaps = [('0001', '0010'), ('0101', '0110'), ('1001', '1010'), ('1101', '1110')]
    assert set(swaps) == set(expected_swaps)
    swaps = _p_automorphism(4, 3, 4)
    expected_swaps = [('0100', '1000'), ('0101', '1001'), ('0110', '1010'), ('0111', '1011')]
    assert set(swaps) == set(expected_swaps)

def test_q_automorphism():
    swaps = _q_automorphism(4, 1, 2)
    expected_swaps = [('0010', '0011'), ('0110', '0111'), ('1010', '1011'), ('1110', '1111')]
    assert set(swaps) == set(expected_swaps)
    swaps = _q_automorphism(4, 3, 4)
    expected_swaps = [('1000', '1100'), ('1001', '1101'), ('1010', '1110'), ('1011', '1111')]
    assert set(swaps) == set(expected_swaps)


class TestQuantumReedMuller:

    def test_swap_type(self, qrms: dict[int, QuantumReedMuller]):
        qrm = qrms[4]
        swap_gates = qrm._automorphism('P', [(1, 2)])._swap_type()
        assert swap_gates == {frozenset({9, 10}), frozenset({13, 14}), frozenset({1, 2}), frozenset({5, 6})}
        swap_gates = qrm._automorphism('P', [(3, 4)])._swap_type()
        assert swap_gates == {frozenset({5, 9}), frozenset({7, 11}), frozenset({6, 10}), frozenset({4, 8})}
        swap_gates = qrm._automorphism('Q', [(1, 2)])._swap_type()
        assert swap_gates == {frozenset({2, 3}), frozenset({6, 7}), frozenset({10, 11}), frozenset({14, 15})}
        swap_gates = qrm._automorphism('Q', [(3, 4)])._swap_type()
        assert swap_gates == {frozenset({11, 15}), frozenset({9, 13}), frozenset({10, 14}), frozenset({8, 12})}

    def test_phase_type(self, qrms: dict[int, QuantumReedMuller]):
        qrm = qrms[4]
        cz_gates, s_gates = qrm._automorphism('P', [(1, 2)])._phase_type()
        assert cz_gates == {frozenset({9, 10}), frozenset({13, 14}), frozenset({1, 2}), frozenset({5, 6})}
        assert s_gates == {0, 3, 12, 15}.union({8, 11, 4, 7})
        cz_gates, s_gates = qrm._automorphism('P', [(3, 4)])._phase_type()
        assert cz_gates == {frozenset({5, 9}), frozenset({7, 11}), frozenset({6, 10}), frozenset({4, 8})}
        assert s_gates == {0, 3, 12, 15}.union({1, 2, 13, 14})
        cz_gates, s_gates = qrm._automorphism('Q', [(1, 2)])._phase_type()
        assert cz_gates == {frozenset({2, 3}), frozenset({6, 7}), frozenset({10, 11}), frozenset({14, 15})}
        assert s_gates == {0, 9, 12, 5}.union({8, 1, 4, 13})
        cz_gates, s_gates = qrm._automorphism('Q', [(3, 4)])._phase_type()
        assert cz_gates == {frozenset({11, 15}), frozenset({9, 13}), frozenset({10, 14}), frozenset({8, 12})}
        assert s_gates == {0, 3, 5, 6}.union({1, 2, 4, 7})


class TestGetIntermediateSubsets:
    
    def test_empty_list(self):
        """Empty list occurs when subsets differ by only one basis vector."""
        b_subset = {1, 2}
        b_prime_subset = {1, 3}
        assert _get_intermediate_subsets(b_subset, b_prime_subset) == []
    
    @pytest.mark.parametrize(
            "b_subset, b_prime_subset",
            [
                ({1, 2}, {3, 4}),
                ({1, 2, 3}, {2, 3, 4}),
                ({1, 2, 3}, {4, 5, 6}),
            ],
    )
    def test_intermediate_subsets(
            self,
            b_subset: set[int],
            b_prime_subset: set[int],
    ):
        subsets = _get_intermediate_subsets(b_subset, b_prime_subset)
        for a, b in itertools.pairwise([b_subset] + subsets + [b_prime_subset]):
            assert len(a.difference(b)) == 1
            assert len(b.difference(a)) == 1


from collections.abc import Collection, Iterable
from itertools import chain, combinations
from typing import Literal

import numpy as np


def complement(universe_size: int, subset: Collection[int], start: int = 1):
    """Return the complement of a subset of m elements.

    :param universe_size: The size m of the universal set.
    :param subset: Subset of the universe.
    :param start: The starting index s of the universal set (default is 1).
    :returns complement_set: The complement within {s, ..., s + m - 1}.
    """
    return set(range(start, start + universe_size)).difference(subset)

# adapted from itertools recipes
def powerset(pairs: Collection[tuple[int, int]], max_cardinality: None | int = None):
    """Iterate subsequences of a set up to an optional maximum cardinality.
    
    Example:
    ``powerset([1,2,3]) → () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)``

    :param pairs: The set (of pairs) to take subsets from.
    :param max_cardinality: Optional cap on subset size.
    :returns chain: An iterator over tuples representing subsets,
        from shortest to longest.
    """
    if max_cardinality is None:
        max_cardinality = len(pairs)
    return chain.from_iterable(combinations(pairs, r) for r in range(max_cardinality+1))


def extract_arguments(index: Literal[0, 1], pairs: Iterable[tuple[int, int]]):
    """Extract the set of arguments at a given position from a pair iterable.

    :param index: Which element of each pair to extract (0 or 1).
    :param pairs: Iterable of integer pairs.
    :returns: A set of extracted integers.
    """
    return {pair[index] for pair in pairs}


def all_bitstrings(length: int) -> tuple[str, ...]:
    """Generate all bitstrings of a given length.

    :param length: Desired bitstring length.
    :returns bitstrings: A tuple of all bitstrings of the given length.
    """
    if length == 0:
        return ('',)
    return tuple(np.binary_repr(k, width=length) for k in range(2**length))


def rref_gf2(matrix: list[np.ndarray[tuple[int], np.dtype[np.bool_]]]):
    """Compute the reduced row echelon form of a binary matrix over GF(2).

    :param matrix: List of row vectors.
    :returns rref: The reduced row echelon form as a numpy array of integers.
    """
    # Note: This pure Python implementation seems to be faster than using the galois package.
    array = np.array(matrix, dtype=np.uint8)
    rows, cols = array.shape
    pivot_row = 0

    for j in range(cols):
        if pivot_row >= rows:
            break

        # Find a pivot (a '1' in the current column)
        i = pivot_row
        while i < rows and array[i, j] == 0:
            i += 1

        if i < rows:
            # Swap rows if the pivot is not in the current pivot_row
            if i != pivot_row:
                array[[i, pivot_row]] = array[[pivot_row, i]]

            # Eliminate other 1s in the current column using XOR (modulo 2 addition)
            for k in range(rows):
                if k != pivot_row and array[k, j] == 1:
                    array[k, :] = array[k, :] ^ array[pivot_row, :]

            pivot_row += 1

    return array


sign_to_power: dict[complex, int] = {1: 0, 1j: 1, -1: 2, -1j: 3}
"""Map from {±1, ±i} to the exponent of i."""


import itertools
from typing import Literal

import pytest
import numpy as np
import stim

from qrmfold.utils import sign_to_power
from qrmfold.quantum_reed_muller import QuantumReedMuller
from qrmfold.utils import rref_gf2


class TestCodespacePreservation:
    """Test the stabilizer group is preserved."""

    def _helper(
            self,
            qrm: QuantumReedMuller,
            circuit: stim.Circuit,
    ):
        new_stabilizers_bsf: list[np.ndarray[tuple[int], np.dtype[np.bool_]]] = []
        for basis_generators in qrm.stabilizer_generators.values():
            for generator in basis_generators:
                morphed = generator.after(circuit)
                assert (sign_to_power[morphed.sign] + str(morphed).count('Y')) % 4 == 0
                xs, zs =  morphed.to_numpy()
                new_stabilizers_bsf.append(np.append(xs, zs))
        new_stabilizer_generators_rref = rref_gf2(new_stabilizers_bsf)
        assert np.array_equal(new_stabilizer_generators_rref, qrm.stabilizer_generators_rref)

    @pytest.mark.parametrize("gate_type", ['swap', 'phase'])
    @pytest.mark.parametrize("automorphism_type", ['P', 'Q'])
    @pytest.mark.parametrize("m", range(2, 12, 2))
    def test_1_pair(
        self,
        m: int,
        automorphism_type: Literal['P', 'Q'],
        gate_type: Literal['swap', 'phase'],
        qrms: dict[int, QuantumReedMuller],
    ):
        """Test U_S(P(i, j)), U_P(P(i, j), U_S(Q(i, j)), U_P(Q(i, j))
        preserve the codespace.
        """
        qrm = qrms[m]
        for k in range(0, m, 2):
            circuit = qrm.automorphism([(k+1, k+2)], automorphism_type, gate_type)
            self._helper(qrm, circuit)

    @pytest.mark.parametrize("gate_type", ['swap', 'phase'])
    @pytest.mark.parametrize("automorphism_type", ['P', 'Q'])
    @pytest.mark.parametrize("m", range(2, 12, 2))
    def test_1_to_maximal_pair_count(
        self,
        m: int,
        automorphism_type: Literal['P', 'Q'],
        gate_type: Literal['swap', 'phase'],
        qrms: dict[int, QuantumReedMuller],
    ):
        """Test U_S(P(K)), U_P(P(K), U_S(Q(K)), U_P(Q(K))
        preserve the codespace for |K| in 1..m/2.
        """
        qrm = qrms[m]
        for pair_count in range(1, m//2 + 1):
            pairs = [(k+1, k+2) for k in range(0, 2*pair_count, 2)]
            circuit = qrm.automorphism(pairs, automorphism_type, gate_type)
            self._helper(qrm, circuit)

    @pytest.mark.parametrize("m", range(2, 12, 2))
    def test_trivial_automorphism_phase_type(self, m: int, qrms: dict[int, QuantumReedMuller]):
        """Test U_P(e) preserves the codespace.
        U_S(e) is identity so need not be tested.
        """
        qrm = qrms[m]
        circuit = qrm.automorphism(gate_type='phase')
        self._helper(qrm, circuit)

    @pytest.mark.parametrize("reduce_depth", (False, True))
    @pytest.mark.parametrize("name", ('ZZCZ', 'SWAP'))
    @pytest.mark.parametrize("m", range(2, 6, 2))
    def test_addressable_gate(
        self,
        m: int,
        name: Literal['ZZCZ', 'SWAP'],
        reduce_depth: bool,
        qrms: dict[int, QuantumReedMuller],
    ):
        qrm = qrms[m]
        for j, k in itertools.combinations(qrm.logical_qubit_ordering.keys(), 2):
            circuit = qrm.gate(name, [j, k], reduce_depth=reduce_depth)
            self._helper(qrm, circuit)


class TestLogicalAction:
    """Test for the correct logical action."""

    @pytest.mark.parametrize(
            "m, pair_count",
            [(m, pair_count) for m in range(2, 12, 2) for pair_count in range(m//2 + 1)],
    )
    def test_single_unitary(self, m: int, pair_count: int, qrms: dict[int, QuantumReedMuller]):
        qrm = qrms[m]
        pairs = [(k+1, k+2) for k in range(0, 2*pair_count, 2)]
        physical_circuit = qrm.automorphism(pairs, 'Q', 'phase')
        realized_tableau = qrm._get_logical_tableau(physical_circuit)
        target_tableau = qrm.q_phase_logical_action(pairs).to_tableau()
        assert realized_tableau == target_tableau

    @pytest.mark.parametrize(
            "m, pair_count",
            [(m, pair_count) for m in range(2, 12, 2) for pair_count in range(m//2 + 1)],
    )
    def test_product_of_unitaries(self, m: int, pair_count: int, qrms: dict[int, QuantumReedMuller]):
        qrm = qrms[m]
        pairs = [(k+1, k+2) for k in range(0, 2*pair_count, 2)]
        physical_circuit = qrm.automorphism_product(pairs, automorphism_type='Q', gate_type='phase')
        realized_tableau = qrm._get_logical_tableau(physical_circuit)
        target_tableau = qrm.q_phase_product_logical_action(pairs).to_tableau()
        assert target_tableau == realized_tableau


class TestAddressableLogicalAction:

    @pytest.mark.parametrize("reduce_depth", (False, True))
    @pytest.mark.parametrize("m", range(2, 8, 2))
    @pytest.mark.parametrize("name", ['S', 'H'])
    def test_1_qubit_gate(
        self,
        qrms: dict[int, QuantumReedMuller],
        m: int,
        name: Literal['S', 'H'],
        reduce_depth: bool,
    ):
        qrm = qrms[m]
        for logical_index in qrm.logical_qubit_ordering.keys():
            physical_circuit = qrm.gate(name, [logical_index], reduce_depth=reduce_depth)
            realized_tableau = qrm._get_logical_tableau(physical_circuit)
            
            target_circuit = stim.Circuit()
            target_circuit.append('I', qrm.logical_qubit_ordering.keys(), ())
            target_circuit.append(name, [logical_index], ())
            target_tableau = target_circuit.to_tableau()
            
            assert target_tableau == realized_tableau

    @pytest.mark.parametrize("m", range(2, 8, 2))
    @pytest.mark.parametrize("name", ['SWAP', 'ZZCZ'])
    def test_restricted_2_qubit_gate(
        self,
        qrms: dict[int, QuantumReedMuller],
        m: int,
        name: Literal['SWAP', 'ZZCZ'],
    ):
        qrm = qrms[m]
        to_test = qrm._zzcz_restricted if name == 'ZZCZ' else qrm._swap_restricted
        gates = ['CZ', 'Z'] if name == 'ZZCZ' else ['SWAP']
        for (i, i_tuple), (j, j_tuple) in itertools.combinations(qrm.logical_qubit_ordering.items(), 2):
            i_subset = set(i_tuple)
            j_subset = set(j_tuple)
            if len(i_subset.intersection(j_subset)) == qrm.BIT_COUNT//2 - 1:
                
                physical_circuit = to_test(i_subset, j_subset)
                realized_tableau = qrm._get_logical_tableau(physical_circuit)
            
                target_circuit = stim.Circuit()
                target_circuit.append('I', qrm.logical_qubit_ordering.keys(), ())
                for stim_gate in gates:
                    target_circuit.append(stim_gate, [i, j], ())
                target_tableau = target_circuit.to_tableau()
            
                assert target_tableau == realized_tableau

    @pytest.mark.parametrize("reduce_depth", (False, True))
    @pytest.mark.parametrize("m", range(2, 8, 2))
    @pytest.mark.parametrize("name", ['SWAP', 'ZZCZ'])
    def test_2_qubit_gate(
        self,
        qrms: dict[int, QuantumReedMuller],
        m: int,
        name: Literal['SWAP', 'ZZCZ'],
        reduce_depth: bool,
    ):
        """Test the logical action of either a SWAP or ZZCZ gate
        on any pair of logical qubits.
        
        Note: there are 4 tests for m = 6 which in total take about 300 s on a laptop.
        """
        qrm = qrms[m]
        gates = ['SWAP'] if name == 'SWAP' else ['CZ', 'Z']
        for i, j in itertools.combinations(qrm.logical_qubit_ordering.keys(), 2):
            physical_circuit = qrm.gate(name, [i, j], reduce_depth=reduce_depth)
            realized_tableau = qrm._get_logical_tableau(physical_circuit)
        
            target_circuit = stim.Circuit()
            target_circuit.append('I', qrm.logical_qubit_ordering.keys(), ())
            for stim_gate in gates:
                target_circuit.append(stim_gate, [i, j], ())
            target_tableau = target_circuit.to_tableau()
        
            assert target_tableau == realized_tableau